/*
 * Copyright 2009-2012 forwardfire.net All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided
 * that the following conditions are met:
 * 
 * * Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *   following disclaimer.
 * * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
 *   the following disclaimer in the documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package net.forwardfire.logstats.ejb.jobs;

import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.logging.Logger;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.annotation.Resource;
import javax.ejb.ActivationConfigProperty;
import javax.ejb.EJB;
import javax.ejb.MessageDriven;
import javax.ejb.TransactionManagement;
import javax.ejb.TransactionManagementType;
import javax.persistence.EntityManager;
import javax.persistence.NoResultException;
import javax.persistence.PersistenceContext;
import javax.transaction.NotSupportedException;
import javax.transaction.SystemException;
import javax.transaction.UserTransaction;

//import org.jboss.annotation.ejb.ResourceAdapter;
//import org.jboss.ejb3.annotation.ResourceAdapter;
import net.forwardfire.logstats.ejb.models.LogData;
import net.forwardfire.logstats.ejb.models.LogHost;
import net.forwardfire.logstats.ejb.models.LogLevel;
import net.forwardfire.logstats.ejb.models.LogServerRun;
import net.forwardfire.logstats.ejb.models.LogSource;
import net.forwardfire.logstats.ejb.models.LogStats;
import net.forwardfire.logstats.ejb.models.LogValueLine;
import net.forwardfire.logstats.ejb.models.LogValueType;

import org.quartz.InterruptableJob;
import org.quartz.Job;
import org.quartz.JobExecutionContext;
import org.quartz.JobExecutionException;
import org.quartz.UnableToInterruptJobException;

import net.forwardfire.vasc.xpql.XPQLParser;
import net.forwardfire.vasc.xpql.ejb3.XpqlQueryManager;
import net.forwardfire.vasc.xpql.query.Query;

/**
 * 1.0 03 Aug 2007
 *
 *	A "Cron-Expression" is a string comprised of 6 or 7 fields separated by white space. The 6 mandatory and 1 optional fields are as follows:
 *		FIELD NAME			ALLOWED VALUES				ALLOWED SPECIAL CHARACTERS
 *		Seconds				0-59						, - * /
 *		Minutes				0-59						, - * /
 *		Hours				0-23						, - * /
 *		Day-of-month		1-31						, - * ? / L W C
 *		Month				1-12 or JAN-DEC				, - * /
 *		Day-of-Week			1-7 or SUN-SAT				, - * ? / L C #
 *		Year (Optional) 	empty, 1970-2099			, - * /
 *		
 *		MORE INFO: http://quartz.sourceforge.net/javadoc/org/quartz/CronTrigger.html
 *
 *
 *
 * @author Willem Cazander
 * @version 1.0 Sep 13, 2007
 */
@MessageDriven(
			messageListenerInterface=Job.class,
			activationConfig = {
		@ActivationConfigProperty(propertyName = "cronTrigger", propertyValue = "13 */2 * * * ?"),
		@ActivationConfigProperty(propertyName = "jobName", propertyValue = "UpdateIndexJob"),
		@ActivationConfigProperty(propertyName = "jobGroup", propertyValue = "system")
})
//@ResourceAdapter("quartz-ra.rar")
@TransactionManagement(TransactionManagementType.BEAN)
public class UpdateIndexJob implements InterruptableJob {

	private Logger logger = Logger.getLogger(UpdateIndexJob.class.getName());

    @EJB(beanInterface=XpqlQueryManager.IRemote.class)
    private XpqlQueryManager xtesManager;
    
	@PersistenceContext
	private EntityManager entityManager;
	
    @Resource
    private UserTransaction utx;
    
    private static boolean running = false;
    
    private int count = 0;
    
    private volatile boolean interrupted = false;
    
	static public Queue<LogData> logDataRealtime = new LinkedBlockingQueue<LogData>();
    
    /**
	 * @see org.quartz.InterruptableJob#interrupt()
	 */
	public void interrupt() throws UnableToInterruptJobException {
		interrupted=true;
		logger.info("Job interrupted");
	}

	public void execute(JobExecutionContext jobExecutionContext) throws JobExecutionException {
        logger.info("executing UpdateIndexJob job");
        long startTime = System.currentTimeMillis();
        
        if (running) {
        	//serverLogManager.logMessage(UpdateIndexJob,"UpdateIndexJob is already running.");
        	return;
        }        
        try {
        	running=true;
                	
        	long startDate = System.currentTimeMillis();
        	long lastId = 0l;
        	try {
        		LogServerRun last = (LogServerRun)xtesManager.executeObject("LogServerRun.getLast", null);
        		lastId = last.getLastDataId();
        	} catch (NoResultException nre) {
        		// first run on this DB
        	}
        	
        	// cache some lists for 5min
        	List<LogValueLine> logValueLines = (List)xtesManager.execute("LogValueLine.getAllIndexActive",null);
        	List<LogHost> logHosts = (List)xtesManager.execute("LogHost.getAll",null);
        	List<LogSource> logSources = (List)xtesManager.execute("LogSource.getAll",null);
        	List<LogValueLineData> lineData = prepareData(logValueLines);
        	logger.info("Cached data sizes; logValueLines: "+logValueLines.size()+" logHosts: "+logHosts.size()+" logSources: "+logSources.size()+" lineData: "+lineData.size() );
        	
        	if (lineData.isEmpty()) {
        		return; // no data to match against
        	}
        	
        	// loop until time is over
        	while (true) {
        		Query q = xtesManager.getQuery("LogData.getFromId");
        		String sql = q.toPreparedSQL(q);
        		javax.persistence.Query q2 = entityManager.createQuery(sql);
        		q2.setParameter(1, lastId);
        		q2.setMaxResults(1000);
        		List<LogData> logData = q2.getResultList();
        		if (logData.isEmpty()==false) {
        			logger.info("Got log data size: "+logData.size());
            		long lastIdNew = parseLogData(logData,lineData,logHosts,logSources);
            		if (lastIdNew>0) {
            			lastId = lastIdNew;
            		}	
        		}
        		if (interrupted) {
        			break;
        		}
            	long thisTime = System.currentTimeMillis();
            	long diffTime = thisTime-startTime;
            	long minTime = 100*1000; // stop working just 20sec before next fire !!
            	if (diffTime>minTime) {
            		break;
            	}
            	Thread.sleep(10*1000); // sleep 10sec before checking again for new records.
        	}

        	LogServerRun thisRun = new LogServerRun();
        	thisRun.setLastDataId(lastId);
        	thisRun.setStartDate(new Date(startDate));
        	thisRun.setStopDate(new Date());
        	
			utx.begin();
			entityManager.persist(thisRun);
			utx.commit();
        	
        } catch (Exception e) {
        	e.printStackTrace();
        	//serverLogManager.logException(e);
        } finally {
        	running=false;
        	long endTime = System.currentTimeMillis();
        	logger.info("UpdateIndexJob done in: "+(endTime-startTime)+" ms updated records: "+count);
        	//serverLogManager.logMessage(UpdateIndexJob.class,"Total UpdateIndexJob updated: "+count+" in "+(endTime-startTime)+" ms");
        }
        
    }
    
    private List<LogValueLineData> prepareData(List<LogValueLine> logValueLines) throws Exception {
    	// create logValueLine List
    	List<LogValueLineData> lines = new ArrayList<LogValueLineData>(logValueLines.size());
    	for (LogValueLine line:logValueLines) {
    		
    		if (line.getActive()==false) {
    			continue;
    		}
    		// todo: check start/stop date
    		
    		// setup data obj
    		LogValueLineData data = new LogValueLineData();
    		data.logStats =  new HashMap<String,LogStats>(100);
    		data.logValueLine=line;
    		
    		String match = line.getMatchRegex();
    		if (match.isEmpty()) {
    			logger.info("Skipped LogValueLine because match regex is empty.");
    			continue;
    		}
    		data.matchPattern = Pattern.compile(match);
    		
    		// fill regex
    		String regex = line.getSplitRegex();
    		if (regex.isEmpty()) {
    			logger.info("Skipped LogValueLine because split regex is empty.");
    			continue;
    		}
    		data.splitPattern = Pattern.compile(regex);
    		
    		lines.add(data);
    	}
    	        	
    	// put already calulated logStats in data obj. 
    	List<LogStats> logStats = (List)xtesManager.execute("LogStats.getLastDay",null);
    	for (int i=0;i<logStats.size();i++) {
    		LogStats data = logStats.get(i);
    		for (LogValueLineData line :lines) {
    			if (line.logValueLine.getLogValue().getId().equals(data.getLogValueId())) {
    				line.logStats.put(data.getValueText(),data);
    				break;
    			}
        	}
    	}
    	
    	for (LogValueLineData line:lines) {
    		logger.info("line: "+line.logValueLine.getId()+" logStats: "+line.logStats.size()+" fetchedStats: "+logStats.size());
    	}
    	
    	return lines;
    }
    
    private long parseLogData(List<LogData> logData,List<LogValueLineData> lines,List<LogHost> logHosts,List<LogSource> logSources) throws Exception {

    	logger.info("parseLogData size: "+logData.size()+" lines: "+lines.size());
    	// process the raw log data
    	Long lastId = null;
    	for (int i=0;i<logData.size();i++) {
    		if (interrupted) {
    			break;
    		}
    		LogData data = logData.get(i);
			logger.info("Check line: "+data.getMessage());
    		lastId = data.getId();
    		for (int l=0;l<lines.size();l++) {
    			LogValueLineData line = lines.get(l);
    			
    			if (line.logValueLine.getLogHost()!=null && line.logValueLine.getLogHost().getHostname().equals(data.getHostname())==false) {
    				continue; // other host
    			}
    			if (line.logValueLine.getLogLevel()!=null && line.logValueLine.getLogLevel().getLevelNumber().equals(data.getLogLevel())==false) {
    				continue; // other level
    			}
    			if (line.logValueLine.getLogSource()!=null && line.logValueLine.getLogSource().getSource().equals(data.getSource())==false) {
    				continue; // other source
    			}
    			
    			Matcher m = line.matchPattern.matcher(data.getMessage());
    			if (m.matches()==false) {
    				continue; // message does not match
    			}
    			
    			
    			String valueType = line.logValueLine.getLogValue().getLogValueType().getValueType();
    			
    			// split
    			String[] columns = line.splitPattern.split(data.getMessage());
    			String value = getString(columns,line.logValueLine.getValueIndexNumber());
    			
    			int timeNumber = 0;
    			if (LogValueType.VALUE_COUNT.equals(valueType)==false) {
    				String time = getString(columns,line.logValueLine.getTimeIndexNumber());
    				timeNumber = parseTime(time);
    			}
    			
    			logger.info("Checking value: '"+value+"' and time: '"+timeNumber+"'");
    			
    			LogStats stats = line.logStats.get(value);
    			if (stats==null) {
    				stats = new LogStats();
    				stats.setLogDay(new Date());
    				stats.setCreatedDate(new Date());
    				stats.setValueText(value);
    				stats.setValueCount(0);
    				stats.setTimeTotal(0);
    				stats.setTimeMin(timeNumber);
    				stats.setTimeMax(timeNumber);
    				stats.setTimeAvg(timeNumber);
    				
    				LogHost logHost = findOrCreateLogHost(logHosts,data.getHostname());
    				stats.setLogHostId(logHost.getId());
    				stats.setLogHost(logHost.getHostname());
    				
    				LogSource logSource = findOrCreateLogSource(logSources,data.getSource());
    				stats.setLogSourceId(logSource.getId());
    				stats.setLogSource(logSource.getSource());
    				
    				stats.setLogValueId(line.logValueLine.getLogValue().getId());
    				stats.setLogValue(line.logValueLine.getLogValue().getName());
    			}
    			stats.setValueCount(stats.getValueCount()+1);
    		
    			// in counting we don't need to calulate
    			if (LogValueType.VALUE_COUNT.equals(valueType)==false) {
    				stats.setTimeTotal(stats.getTimeTotal()+timeNumber);
    				if (stats.getTimeMin()>timeNumber) {
    					stats.setTimeMin(timeNumber);
    				}
    				if (stats.getTimeMax()<timeNumber) {
    					stats.setTimeMax(timeNumber);
    				}
    				stats.setTimeAvg(stats.getTimeTotal()/stats.getValueCount());
    			}
    			
    			// mark changed
    			stats.setModifiedDate(new Date(0));
    			
    			//put in map
    			line.logStats.put(stats.getValueText(), stats);
    		}
    		
    		// data is processed lets kick to realtime view for last usage of the obj.
    		logDataRealtime.add(data);
    		if (logDataRealtime.size()>100) {
    			logDataRealtime.remove();
    		}
    	}

    	// save changed stats records.
    	for (LogValueLineData line:lines) {
    		utx.begin();
			for (String key:line.logStats.keySet()) {
				LogStats stats = line.logStats.get(key);
				if (stats.getModifiedDate().getTime()==0l) {
					stats.setModifiedDate(new Date());
					entityManager.merge(stats);	    			
					count++;
				}
			}
    		utx.commit();
    	}
    	
    	return lastId;
    }
    
    private LogHost findOrCreateLogHost(List<LogHost> logHosts,String hostname) throws Exception {
    	for (int i=0;i<logHosts.size();i++) {
    		LogHost logHost = logHosts.get(i);
    		if (logHost.getHostname().equalsIgnoreCase(hostname)) {
    			return logHost;
    		}
    	}
    	// not found lets create
    	LogHost logHost = new LogHost();
    	logHost.setActive(true);
    	logHost.setHostname(hostname);
    	
		utx.begin();
		logHost = entityManager.merge(logHost);
		utx.commit();
		
		logHosts.add(logHost);
		
		return logHost;
    }
    
    private LogSource findOrCreateLogSource(List<LogSource> logSources,String source) throws Exception {
    	for (int i=0;i<logSources.size();i++) {
    		LogSource logSource = logSources.get(i);
    		if (logSource.getSource().equalsIgnoreCase(source)) {
    			return logSource;
    		}
    	}
    	// not found lets create
    	LogSource logSource = new LogSource();
    	logSource.setActive(true);
    	logSource.setSource(source);
    	
		utx.begin();
		logSource = entityManager.merge(logSource);
		utx.commit();
		
		logSources.add(logSource);
		
		return logSource;
    }
    
    private int parseTime(String valueTime) {
    	int msIndex = valueTime.indexOf("ms");
    	if (msIndex>0) {
    		valueTime=valueTime.substring(0, msIndex); // remove "ms" from Long
    	}
    	return new Integer(valueTime);
    }
    
    private String getString(String[] lineValues,int index) {
        try {
        	return lineValues[index];
        } catch (ArrayIndexOutOfBoundsException e) {
        	return "";
        }
    }

    
    class LogValueLineData {
    	LogValueLine logValueLine = null;
    	Pattern matchPattern = null;
    	Pattern splitPattern = null;
    	Map<String,LogStats> logStats = null;
    }
}
