/*
 * Copyright 2009-2012 forwardfire.net All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without modification, are permitted provided
 * that the following conditions are met:
 * 
 * * Redistributions of source code must retain the above copyright notice, this list of conditions and the
 *   following disclaimer.
 * * Redistributions in binary form must reproduce the above copyright notice, this list of conditions and
 *   the following disclaimer in the documentation and/or other materials provided with the distribution.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL
 * THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT
 * OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
 * SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

package net.forwardfire.logstats.ejb.models;

import java.io.Serializable;
import java.lang.reflect.Method;
import java.util.HashMap;
import java.util.Map;

import javax.persistence.Id;

/**
 * Base class for all enities
 * 
 * Currenly only implements the overriden equals method.
 * On list to do;
 * - hashCode
 * - 3x date
 * 
 * @author Willem Cazander
 * @version 1.0 Aug 08, 2009
 */
abstract public class AbstractEntityBase implements Serializable {
	
	/** The serial version ID */
	private static final long serialVersionUID = 10l;
	
	/**
	 * static cacheing so it is only scanned one per classloader
	 */
	static private Map<Class<?>,Method> cachedEntityId = new HashMap<Class<?>,Method>(100);
	
	/**
	 * Scannes the class for the "@Id" annotation and caches the result.
	 * 
	 * @param clazz	The class the get the id method for.
	 * @return	The Method to invoke to return the Id of the clazz.
	 */
	static private Method lookupEntityIdMethod(Class<?> clazz) {
		Method entityId = cachedEntityId.get(clazz);
		if (entityId!=null) {
			return entityId;
		}
		
		// scan class
		Method[] methods = clazz.getMethods();
		for (Method method : methods) {
			if (method.isAnnotationPresent(Id.class)) {
				entityId = method;
				break;
			}
		}
		if (entityId==null) {
			throw new RuntimeException("Can't equal/hash entity bean with no @Id annotation class: "+clazz);
		}
		cachedEntityId.put(clazz,entityId);
		return entityId;
	}
	
	/**
	 * @see java.lang.Object#equals(java.lang.Object)
	 */
	@Override
	public boolean equals(Object obj) {
		if ( obj == null ) {
			return false;  // false on null
		}
		if ( this.hashCode() == obj.hashCode() ) {
			return true;   // true on same object ?
		}
		if ( obj.getClass() != this.getClass() ) {
			return false;  // false on other classes incl subclasses !
		}
		
		Method entityId = lookupEntityIdMethod(this.getClass());
		Object idThis = null;
		Object idObj = null;
		try {
			idThis = entityId.invoke(this, (Object[]) null);
			idObj  = entityId.invoke(obj,  (Object[]) null);
		} catch (Exception e) {
			throw new RuntimeException(e);
		}
		if (idThis==null | idObj==null) {
			return super.equals(obj); // no id, do super
		}
		boolean result = idThis.equals(idObj);
		return result;
	}
}