

package net.forwardfire.logstats.jboss.log4j;

import java.net.InetAddress;
import java.net.UnknownHostException;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.Timestamp;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.sql.DataSource;

import org.apache.log4j.Level;
import org.apache.log4j.spi.LoggingEvent;
import org.jboss.logging.Logger;

/**
 * Perists the messages of the queue
 * 
 * @author Willem Cazander
 * @version 1.0 Jun 09, 2010
 */
public class DataSourceLogPersistThread extends Thread {

	private DataSourceLogService logService = null;
	private Logger logger = Logger.getLogger(DataSourceLogPersistThread.class.getName());

	/**
	 * Constructor with the LogService
	 * @param logService
	 */
	public DataSourceLogPersistThread(DataSourceLogService logService) {
	    this.logService = logService;
	}

	/**
	 * Perists the messages of the queue
	 * 
	 * @see java.lang.Thread#run()
	 */
	@Override
	public void run() {
		long idleTimer = System.currentTimeMillis();
		
		// setup DB connection from datasource
		Connection conn = null;
		try {
			conn = getConnection();
		} catch (SQLException e) {
			logger.error("init getConnection failed, error: "+e.getMessage(),e);
			return;
		} catch (NamingException e) {
			logger.error("init getConnection datasource lookup failed, error: "+e.getMessage(),e);
			return;
		}
		
		// Get hostname setup
		String hostname = null;
		try {
			hostname = InetAddress.getLocalHost().getHostName();
		} catch (UnknownHostException e) {
			logger.error("init getLocalHost failed, error: "+e.getMessage(),e);
			try {
				conn.close();
			} catch (SQLException se) {
				logger.warn("init connection close failed, SQLException: "+se.getMessage(),e);
			}
			return;
		}
		if (logService.getBackendHostname()!=null && logService.getBackendHostname().isEmpty()==false) {
			hostname = logService.getBackendHostname();
		}
		
		// Test and record out one thread
		try {
			storeLogData(conn,hostname,System.currentTimeMillis(),Level.DEBUG.toInt(),this.getClass().getName(),"Startup and connection of thread: "+getName());
		} catch (SQLException e) {
			logger.error("test connection message error: "+e.getMessage(),e);
			try {
				if (conn!=null && conn.isClosed()==false) {
					conn.close();
				}
			} catch (SQLException se) {
				logger.warn("test connection close failed, SQLException: "+se.getMessage(),se);
			}
			return;
		}
		
		// main loop
		int errorCount = 0;
		while (isInterrupted()==false) {
			LoggingEvent logMessage = null;
			try {
				logMessage = logService.pollLoggingEvent();
			} catch (InterruptedException e) {
				break;
			} 
			if (logMessage!=null) {
				try {
					// convert
					String message = convertMessage(logMessage.getMessage());
					
					// persist message
					storeLogData( conn,hostname,logMessage.timeStamp,logMessage.getLevel().toInt(),logMessage.getLoggerName(),message );
					
				} catch (SQLException se) {
					if (errorCount==0) {
						logger.error("storeLogData() failed, error: "+se.getMessage());
					}
					errorCount++;
					if (errorCount>1000) {
						logger.error("storeLogData() failed 1000 times, last error: "+se.getMessage());
						errorCount=0;
					}
				}
			} else {
				if (checkConnection(conn) == false) {
					try {
						if (conn!=null && conn.isClosed()==false) {
							conn.close();
						}
					} catch (SQLException se) {
						logger.warn("connection close failed, SQLException: "+se.getMessage(),se);
					}
					try {
						conn = getConnection();
					} catch (Exception ee) {
						logger.error("failure in getConnection() error:"+ee.getMessage(),ee);
						break;
					}
				} else {
					if (logService.getBackendIdleMarkerTime()>0) {
						// check for update of idle timer
						if ((System.currentTimeMillis()-idleTimer) > logService.getBackendIdleMarkerTime()*1000) {
							idleTimer = System.currentTimeMillis();
							logger.info(logService.getBackendIdleMarkerMessage()+" thread: "+getName());
						}
					}
				}
			}
		}
		
		// exit thread
		logger.info("shutdown thread: "+getName());
		try {
			if (conn!=null && conn.isClosed()==false) {
				conn.close();
			}
		} catch (SQLException e) {
			logger.warn("connection close failed, SQLException: "+e.getMessage());
		}
	}
	
	private Connection getConnection() throws SQLException,NamingException {
		String dataSourceName = logService.getBackendDataSource();
		logger.debug("Getting connection from datasource: "+dataSourceName+" for thread: "+getName());
		Context naming = new InitialContext();
		DataSource dataSource = (DataSource) naming.lookup(dataSourceName);
		return dataSource.getConnection();
	}

	public boolean checkConnection(Connection connection) {
		if (connection==null) {
			return false;
		}
		try {
			return connection.prepareStatement(logService.getBackendKeepAliveQuery()).execute();
		} catch (SQLException e) {
			logger.error("checkConnection: SQLException: "+e.getMessage(),e);
			return false;
		}
	}

	/**
	 * Message is not always a string so check for some types we can handle.
	 * 
	 * @param messageObj
	 * @return
	 */
	private String convertMessage(Object messageObj) { 
		String message = null;
		if (messageObj == null ) {
			message = "NULL";
		} else if (messageObj instanceof String) {
			message = (String)messageObj;
		} else if (messageObj instanceof Exception) {
			message = ((Exception)messageObj).getMessage();
		} else {
			message = messageObj.toString();
		}
		return message;
	}
	
	private void storeLogData(Connection connection,String host,long time, int level, String source, String message) throws SQLException {
		PreparedStatement stmt = null;
		Timestamp timeStamp = new Timestamp(time);
		stmt = connection.prepareStatement(logService.getBackendLogQuery());
		String para = logService.getBackendLogQueryParameters();
		String[] paras = para.split(",");
		for (int i=1;i<=paras.length;i++) {
			String col = paras[i-1];
			if ("time".equals(col)) {
				stmt.setTimestamp(i, timeStamp);	  
			} else if ("level".equals(col)) {
				stmt.setInt(i, level);	  
			} else if ("host".equals(col)) {
				stmt.setString(i, host);
			} else if ("source".equals(col)) {
				stmt.setString(i, source);
			} else if ("message".equals(col)) {
				stmt.setString(i, message);
			} else {
				throw new IllegalArgumentException("Unknow column: "+col+" on index: "+i);
			}
		}
		stmt.execute();
	}
}