package net.forwardfire.logstats.jboss.log4j;

import java.io.InputStream;
import java.util.ArrayList;
import java.util.List;
import java.util.concurrent.BlockingQueue;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.TimeUnit;

import javax.xml.parsers.FactoryConfigurationError;

import org.apache.log4j.LogManager;
import org.apache.log4j.Logger;
import org.apache.log4j.spi.LoggingEvent;
import org.apache.log4j.xml.DOMConfigurator;

import org.jboss.logging.util.OnlyOnceErrorHandler;
import org.jboss.system.ServiceMBeanSupport;

/**
 * Jboss MBean which hooks DB logger to log4j and depends on a datasource for db access.
 * 
 * Orginal idea was made by Edward Pilipczuk <exedx@aster.pl>
 * 
 * @author Willem Cazander
 * @version 1.0 Jun 09, 2010
 */
public class DataSourceLogService extends ServiceMBeanSupport implements DataSourceLogServiceMBean {

	// backend config options
	private String	backendDataSource = null;
	private String	backendLogQuery  = "INSERT INTO log_data (log_date, log_level,hostname, source, message) VALUES (?,?,?,?,?)";
	private String	backendLogQueryParameters = "time,level,host,source,message";
	private String	backendHostname = null;
	private int		backendThreads = 3;
	private int		backendMaxQueueSize = 1500;
	private int		backendKeepAliveTimeout = 300;
	private String	backendKeepAliveQuery = "SELECT 1";
	private int		backendIdleMarkerTime = 3600;
	private String	backendIdleMarkerMessage = "--- Mark ---";
	private boolean backendStatisticsLog = true;
	private int		backendStatisticsLogTime = 300;

	// log4j config options
	private String	log4jParseConfigResource = "META-INF/logstats-log4j.xml";
	private String	log4jAppenderName = "LOGSTATS";
	private boolean	log4jAppenderAutoAdd = true;
	private boolean log4jAppenderAutoAddOnlyOnceError = true;
	private boolean log4jAppenderToRoot = true;

	// stats counters
	private int		statsMessagesTotal = 0;
	private int		statsMessagesDropped = 0;
	
	// Internal state objects
	volatile private boolean running = false;
	private BlockingQueue<LoggingEvent> messageQueue = null;
	private DataSourceLog4jAppender appender = null; 
	private List<DataSourceLogPersistThread> persistThreads = null;

	// ===== Override ServiceMBeanSupport methods

	@Override
	protected void startService() throws Exception {
		messageQueue = new LinkedBlockingQueue<LoggingEvent>(backendMaxQueueSize);
		startLog4j();
		startBackend();
		running = true;
		log.info("DataSourceLogService started; queueSize: "+backendMaxQueueSize+" threads: "+backendThreads);
	}

	@Override
	protected void stopService() throws Exception {
		running = false;
		try {
			stopLog4j();
		} catch (Exception e) {
			log.warn("Error in stopLog4j: "+e.getMessage());
		} finally {
			try {
				stopBackend();
			} catch (Exception e) {
				log.warn("Error in stopBackend: "+e.getMessage());
			} finally {
				log.info("DataSourceLogService stopped.");
			}
		}
	}
	
	// =========== Private methods
	    
	private void startBackend() {
		persistThreads = new ArrayList<DataSourceLogPersistThread>(backendThreads); 
		for (int i=0;i<backendThreads;i++) {
			DataSourceLogPersistThread worker = new DataSourceLogPersistThread(this);
			worker.start();
			persistThreads.add(worker);
		}
	}
	
	private void stopBackend() {
		for (DataSourceLogPersistThread worker:persistThreads) {
			worker.interrupt();
		}
		try {
			Thread.sleep(500);
		} catch (InterruptedException e) {
			log.warn("Interrupted while wait until backend threads are done.");
		}
		persistThreads.clear();
		messageQueue.clear();
	}
	
	private void startLog4j() {
		appender = new DataSourceLog4jAppender(this);		
		if (log4jAppenderAutoAddOnlyOnceError) {
			appender.setErrorHandler(new OnlyOnceErrorHandler());
			log.debug("Added OnlyOnceErrorHandler to appender: "+appender.getName());
		}
				
		if (log4jAppenderAutoAdd) {
			Logger.getRootLogger().addAppender(appender);
			log.debug("Added appender: "+appender.getName()+" to root logger.");
		}
		
		if (log4jParseConfigResource!=null && log4jParseConfigResource.isEmpty()==false) {
			try {
				InputStream lis = Thread.currentThread().getContextClassLoader().getResourceAsStream(log4jParseConfigResource);
				new DOMConfigurator().doConfigure(lis,LogManager.getLoggerRepository());
				log.debug("Loaded config for log4j: "+log4jParseConfigResource);
			} catch (FactoryConfigurationError e) {
				log.warn("Error while configing log4j from "+log4jParseConfigResource+" error: "+e.getMessage());
			}
		}
	}
	
	private void stopLog4j() {
		Logger.getRootLogger().removeAppender(appender);
		appender.shutdown();
		appender = null;
	}

	// ============== Protected methods
	
	protected boolean isRunning() {
		return running;
	}
	
	protected void putLoggingEvent(LoggingEvent event) throws InterruptedException {
		if (running==false) {
			return;
		}
		if (messageQueue==null) {
			throw new IllegalStateException("LogService is not started");
		}
		statsMessagesTotal++;
		if (messageQueue.remainingCapacity()>0) {
			messageQueue.put(event);
		} else {
			statsMessagesDropped++;
		}
	}
	
	protected LoggingEvent pollLoggingEvent() throws InterruptedException {
		if (messageQueue==null) {
			throw new IllegalStateException("LogService is not started");
		}
		return messageQueue.poll(backendKeepAliveTimeout, TimeUnit.SECONDS);
	}
	
	// ============== MBean methods
	
	/**
	 * Resets the stats counters
	 */
	public void resetStatsCounters() {
		statsMessagesTotal   = 0;
		statsMessagesDropped = 0;
	}
	
	/**
	 * @return the statsMessagesTotal
	 */
	public int getStatsMessagesTotal() {
		return statsMessagesTotal;
	}

	/**
	 * @return the statsMessagesDropped
	 */
	public int getStatsMessagesDropped() {
		return statsMessagesDropped;
	}
	
	/**
	 * @return the backendDataSource
	 */
	public String getBackendDataSource() {
		return backendDataSource;
	}
	
	/**
	 * @param backendDataSource the backendDataSource to set
	 */
	public void setBackendDataSource(String backendDataSource) {
		this.backendDataSource = backendDataSource;
	}

	/**
	 * @return the backendLogQuery
	 */
	public String getBackendLogQuery() {
		return backendLogQuery;
	}

	/**
	 * @param backendLogQuery the backendLogQuery to set
	 */
	public void setBackendLogQuery(String backendLogQuery) {
		this.backendLogQuery = backendLogQuery;
	}
	
	/**
	 * @return the backendLogQueryParameters
	 */
	public String getBackendLogQueryParameters() {
		return backendLogQueryParameters;
	}

	/**
	 * @param backendLogQueryParameters the backendLogQueryParameters to set
	 */
	public void setBackendLogQueryParameters(String backendLogQueryParameters) {
		this.backendLogQueryParameters = backendLogQueryParameters;
	}

	/**
	 * @return the backendHostname
	 */
	public String getBackendHostname() {
		return backendHostname;
	}

	/**
	 * @param backendHostname the backendHostname to set
	 */
	public void setBackendHostname(String backendHostname) {
		this.backendHostname = backendHostname;
	}

	/**
	 * @return the backendKeepAliveTimeout
	 */
	public int getBackendKeepAliveTimeout() {
		return backendKeepAliveTimeout;
	}

	/**
	 * @param backendKeepAliveTimeout the backendKeepAliveTimeout to set
	 */
	public void setBackendKeepAliveTimeout(int backendKeepAliveTimeout) {
		this.backendKeepAliveTimeout = backendKeepAliveTimeout;
	}

	/**
	 * @return the backendKeepAliveQuery
	 */
	public String getBackendKeepAliveQuery() {
		return backendKeepAliveQuery;
	}

	/**
	 * @param backendKeepAliveQuery the backendKeepAliveQuery to set
	 */
	public void setBackendKeepAliveQuery(String backendKeepAliveQuery) {
		this.backendKeepAliveQuery = backendKeepAliveQuery;
	}

	/**
	 * @return the backendThreads
	 */
	public int getBackendThreads() {
		return backendThreads;
	}

	/**
	 * @param backendThreads the backendThreads to set
	 */
	public void setBackendThreads(int backendThreads) {
		this.backendThreads = backendThreads;
	}

	/**
	 * @return the backendMaxQueueSize
	 */
	public int getBackendMaxQueueSize() {
		return backendMaxQueueSize;
	}

	/**
	 * @param backendMaxQueueSize the backendMaxQueueSize to set
	 */
	public void setBackendMaxQueueSize(int backendMaxQueueSize) {
		this.backendMaxQueueSize = backendMaxQueueSize;
	}

	/**
	 * @return the backendIdleMarkerTime
	 */
	public int getBackendIdleMarkerTime() {
		return backendIdleMarkerTime;
	}

	/**
	 * @param backendIdleMarkerTime the backendIdleMarkerTime to set
	 */
	public void setBackendIdleMarkerTime(int backendIdleMarkerTime) {
		this.backendIdleMarkerTime = backendIdleMarkerTime;
	}

	/**
	 * @return the backendIdleMarkerMessage
	 */
	public String getBackendIdleMarkerMessage() {
		return backendIdleMarkerMessage;
	}

	/**
	 * @param backendIdleMarkerMessage the backendIdleMarkerMessage to set
	 */
	public void setBackendIdleMarkerMessage(String backendIdleMarkerMessage) {
		this.backendIdleMarkerMessage = backendIdleMarkerMessage;
	}

	/**
	 * @return the backendStatisticsLog
	 */
	public boolean isBackendStatisticsLog() {
		return backendStatisticsLog;
	}

	/**
	 * @param backendStatisticsLog the backendStatisticsLog to set
	 */
	public void setBackendStatisticsLog(boolean backendStatisticsLog) {
		this.backendStatisticsLog = backendStatisticsLog;
	}

	/**
	 * @return the backendStatisticsLogTime
	 */
	public int getBackendStatisticsLogTime() {
		return backendStatisticsLogTime;
	}

	/**
	 * @param backendStatisticsLogTime the backendStatisticsLogTime to set
	 */
	public void setBackendStatisticsLogTime(int backendStatisticsLogTime) {
		this.backendStatisticsLogTime = backendStatisticsLogTime;
	}

	/**
	 * @return the log4jParseConfigResource
	 */
	public String getLog4jParseConfigResource() {
		return log4jParseConfigResource;
	}

	/**
	 * @param log4jParseConfigResource the log4jParseConfigResource to set
	 */
	public void setLog4jParseConfigResource(String log4jParseConfigResource) {
		this.log4jParseConfigResource = log4jParseConfigResource;
	}

	/**
	 * @return the log4jAppenderName
	 */
	public String getLog4jAppenderName() {
		return log4jAppenderName;
	}

	/**
	 * @param log4jAppenderName the log4jAppenderName to set
	 */
	public void setLog4jAppenderName(String log4jAppenderName) {
		this.log4jAppenderName = log4jAppenderName;
	}

	/**
	 * @return the log4jAppenderAutoAdd
	 */
	public boolean isLog4jAppenderAutoAdd() {
		return log4jAppenderAutoAdd;
	}

	/**
	 * @param log4jAppenderAutoAdd the log4jAppenderAutoAdd to set
	 */
	public void setLog4jAppenderAutoAdd(boolean log4jAppenderAutoAdd) {
		this.log4jAppenderAutoAdd = log4jAppenderAutoAdd;
	}

	/**
	 * @return the log4jAppenderAutoAddOnlyOnceError
	 */
	public boolean isLog4jAppenderAutoAddOnlyOnceError() {
		return log4jAppenderAutoAddOnlyOnceError;
	}

	/**
	 * @param log4jAppenderAutoAddOnlyOnceError the log4jAppenderAutoAddOnlyOnceError to set
	 */
	public void setLog4jAppenderAutoAddOnlyOnceError(
			boolean log4jAppenderAutoAddOnlyOnceError) {
		this.log4jAppenderAutoAddOnlyOnceError = log4jAppenderAutoAddOnlyOnceError;
	}

	/**
	 * @return the log4jAppenderToRoot
	 */
	public boolean isLog4jAppenderToRoot() {
		return log4jAppenderToRoot;
	}

	/**
	 * @param log4jAppenderToRoot the log4jAppenderToRoot to set
	 */
	public void setLog4jAppenderToRoot(boolean log4jAppenderToRoot) {
		this.log4jAppenderToRoot = log4jAppenderToRoot;
	}
}